import sys, os, copy, logging
from urllib.parse import urlparse
import yt_dlp
from colorboy import cyan, green, red
from deep_filter import deep_filter
from shlex import quote

from vidl import log, config, md as md_module

def parse_cli_options():

    options = {
        'url': '',
        'file_format': 'mp3',
        'audio_only': True,
        'no_md': False,
        'no_thumbnail_embed': False,
        'no_smart_md': False,
        'no_dl': False,
        'verbose': False,
        'download_folder': config.get_config('download_folder'),
        'output_template': config.get_config('output_template'),
    }
    video_formats = ['mp4']
    audio_formats = ['mp3', 'wav', 'm4a']

    # parse arguments
    for arg in sys.argv[1:]:
        if arg in audio_formats:
            options['audio_only'] = True
            options['file_format'] = arg
        elif arg in video_formats:
            options['audio_only'] = False
            options['file_format'] = arg
        elif arg in ['--no-md']:
            options['no_md'] = True
        elif arg in ['--no-smart-md']:
            options['no_smart_md'] = True
        elif arg in ['--no-dl']:
            options['no_dl'] = True
        elif arg in ['--no-embed']:
            options['no_thumbnail_embed'] = True
        elif arg in ['-v', '--verbose']:
            options['verbose'] = True
        elif '.' in arg:
            parsed_url = urlparse(arg)
            url = arg
            if (parsed_url.scheme == ''): url = 'https://'+url
            print(url)
            options['url'] = url
        else:
            log.fatal('Unknown argument:', arg)
    if 'url' == '':
        log.fatal('No URL provided')

    return options

def download(options):
    """Accepts an `options` dict, but there's no validation and all options must be present. Look inside `parse_cli_options()` for an example options object."""

    id3_metadata_formats = ['mp3']
    ytdl_output_template = os.path.join(options['download_folder'], options['output_template'])


    # get info
    log('Fetching URL info')
    ytdl_get_info_options = {
        'outtmpl': ytdl_output_template,
        'quiet': False if options['verbose'] else True,
    }
    with yt_dlp.YoutubeDL(ytdl_get_info_options) as ytdl:
        try:
            info_result = ytdl.extract_info(options['url'], download=False)
        except Exception as err:
            if options['verbose']: logging.exception(err)
            log.fatal('youtube-dl failed to get URL info')
        if options['verbose']: log.pretty(info_result)

    # delete None properties/indexes
    def callback(value):
        return value != None
    cleaned_info_result = deep_filter(copy.deepcopy(info_result), callback)

    # restructure
    url_info = copy.deepcopy(cleaned_info_result)
    if 'entries' in cleaned_info_result:
        videos = cleaned_info_result['entries']
        playlist_info = copy.deepcopy(cleaned_info_result)
        del playlist_info['entries']
    else:
        videos = [cleaned_info_result]
        playlist_info = {}

    # generate ytdl arguments
    ytdl_args = []
    if options['audio_only']:
        ytdl_args += ['-x']
        # yt-dlp specific: best audio, even if it has video
        ytdl_args += ['-f', 'ba*']
        ytdl_args += ['--format-sort', 'abr,codec']
        # in youtube-dl, "--audio-format mp3 --audio-quality 0" seems to not work
        ytdl_args += ['--audio-format', options['file_format']]
        ytdl_args += ['--audio-quality', '0']
    else:
        # yt-dlp specific: best video, and add audio if it's not there
        ytdl_args += ['-f', 'bv*+ba/b']
        ytdl_args += ['--recode-video', options['file_format']]
    ytdl_args += ['-o', ytdl_output_template]
    if options['file_format'] in ['mp3', 'm4a', 'mp4'] and options['no_thumbnail_embed'] == False:
        ytdl_args += ['--embed-thumbnail']
    if not options['verbose']:
        ytdl_args += ['--quiet']
    # band-aid solution for 403 error caused by cache
    ytdl_args += ['--rm-cache-dir']

    video_index = -1
    first_video_artist = ''
    errors = []
    for video in videos:
        video_index += 1
        try:
            filename = ytdl.prepare_filename(video)
        except (Exception, SystemExit) as err:
            if options['verbose']: logging.exception(err)
            error_msg = 'Failed to generate a filename for URL: '+green(video['webpage_url'])
            if len(videos) == 1: log.fatal(error_msg)
            log.error(error_msg)
            errors.append(video)
            continue
        filename_split = filename.split('.')
        filename_split[len(filename_split)-1] = options['file_format']
        filename = '.'.join(filename_split)
        if options['verbose']:
            # print youtube-dl command:
            command = green('youtube-dl command: ')+'youtube-dl '
            for arg in ytdl_args+[video['webpage_url']]:
                command += quote(arg)+' '
            log(command)
        if options['no_dl']:
            continue
        log('Downloading')

        # download
        try:
            yt_dlp.main(ytdl_args+[video['webpage_url']])
        except (Exception, SystemExit) as err:
            if type(err) == SystemExit and err.code == 0:
                # don't treat sys.exit(0) as error
                pass
            else:
                if options['verbose']: logging.exception(err)
                error_msg = 'Failed to download URL: '+green(video['webpage_url'])
                if len(videos) == 1: log.fatal(error_msg)
                log.error(error_msg)
                errors.append(video)
                continue
        log('Saved as', filename)

        # id3 tags
        if options['file_format'] in id3_metadata_formats and not options['no_md']:
            log('Adding metadata to file')

            # get artist/title from title
            parsed_title = {}
            if not options['no_smart_md']:
                if 'title' in video and video['title'].count(' - ') == 1:
                    split_title = video['title'].split(' - ')
                    parsed_title['artist'] = split_title[0]
                    parsed_title['title'] = split_title[1]

            md = {}
            playlist = True if len(videos) > 1 else False

            # smart title
            if 'title' in parsed_title:
                smart_title = True
            else:
                smart_title = False
            # title
            if 'title' in video:
                md['title'] = video['title']
            elif 'track' in video:
                md['title'] = video['track']

            # smart artist
            if 'artist' in parsed_title:
                smart_artist = True
            else:
                smart_artist = False
            # artist
            if 'uploader' in video:
                md['artist'] = video['uploader']
            elif 'artist' in video:
                md['artist'] = video['artist']
            # youtube music artist
            if  video['extractor'] == 'youtube' \
            and video['uploader'].endswith(' - Topic') \
            and 'artist' in video:
                if 'categories' not in video:
                    md['artist'] = video['artist']
                elif video['categories'] == ['Music']:
                    md['artist'] = video['artist']

            use_first_video_artist = False
            if playlist:
                #album
                if 'title' in playlist_info:
                    md['album'] = playlist_info['title']
                elif 'playlist_title' in video:
                    md['album'] = video['playlist_title']
                elif 'playlist' in video and type(video['playlist']) == str:
                    md['album'] = video['playlist']
                #album_artist
                if 'uploader' in playlist_info:
                    md['album_artist'] = playlist_info['uploader']
                elif 'playlist_uploader' in video:
                    md['album_artist'] = video['playlist_uploader']
                else:
                    use_first_video_artist = True
                # track_number
                if 'playlist_index' in video:
                    md['track_number'] = video['playlist_index']
                else:
                    md['track_number'] = video_index+1
                # track_count
                if 'n_entries' in video:
                    md['track_count'] = video['n_entries']
                else:
                    md['track_count'] = len(videos)
            # year
            def is_int(number):
                try:
                    int(number)
                    return True
                except ValueError:
                    return False
            if 'release_date' in video and is_int(video['release_date'][:4]):
                md['year'] = video['release_date'][:4]
            elif 'publish_date' in video and is_int(video['publish_date'][:4]):
                md['year'] = video['publish_date'][:4]
            elif 'upload_date' in video and is_int(video['upload_date'][:4]):
                md['year'] = video['upload_date'][:4]
            
            dumb_md = copy.deepcopy(md)
            if smart_title: md['title'] = parsed_title['title']
            if smart_artist: md['artist'] = parsed_title['artist']

            if playlist:
                # save artist of first video

                if video_index == 0 and 'artist' in md:
                    first_video_artist = md['artist']

                # use first video's artist as album artist if no other is found
                if use_first_video_artist:
                    md['album_artist'] = first_video_artist

            md = config.user_md_parser(md, dumb_md, video, url_info)
            
            md_module.add_metadata(filename, md)
    
    if len(errors) >= 1:
        msg = 'There were errors when downloading the following URLs:'
        for video in errors:
            msg += f'\n- {green(video["webpage_url"])}: {video["uploader"]} - {video["title"]}'
        log.fatal(msg)
    log('Done')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vidl']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'colorboy>=1.0.3,<2.0.0',
 'deep-filter>=1.0.1,<2.0.0',
 'mutagen>=1.45.1,<2.0.0',
 'yt-dlp>2021.0.0']

entry_points = \
{'console_scripts': ['vidl = vidl.run:main']}

setup_kwargs = {
    'name': 'vidl',
    'version': '3.6.3',
    'description': 'Downloads video/audio from anywhere, adding thumbnails and mp3 metadata if applicable',
    'long_description': '<p align="center">\n  <img src="https://raw.githubusercontent.com/probablykasper/vidl/master/logo/logo%201000.png" width="100">\n</p>\n\n# vidl\nvidl is a script designed to easily download video/audio from anywere, using youtube-dl. It automatically embeds thumbnails to mp3/mp4/m4a files.\n\nvidl will add metadata to mp3 files if it\'s found. The `--no-md` option turns this off.\n`title`, `artist` and `year` metadata is added, but if the URL is a playlist, it also adds `album`, `album artist`, `track number`, `track count`.\nIf the title contains " - ", vidl often uses what comes before and after it as artist and title respectively. The `--dont-extract-md` option turns off this behaviour.\n\n# Installation\n1. Install Python (3.7+ is recommended)\n2. Install [ffmpeg and ffprobe](https://www.ffmpeg.org/)\n3. Run `pip3 install vidl`.\n4. If you\'re not on macOS or Windows, you need to specify where vidl will download files to in your vidl config file. Run `vidl --config-path` to see where the config file is.\nIf you\'re on macOS, I recommend [setting up shortcuts for vidl](#macos-shortcut-setup)\n\n# Updating\nvidl is updated by running `pip3 install vidl --upgrade --upgrade-strategy eager`.\nIf vidl is unable to download a URL, it might be because youtube-dl is outdated. The `--upgrade-strategy eager` part updates youtube-dl.\n\n# Uninstall\nRun `pip3 uninstall vidl`.\n\nTo fully uninstall vidl, go to vidl\'s config file (Run `vidl --config-path` to see where it is) and delete the folder it\'s in.\n\n# Usage\nExamples:\n`vidl https://www.youtube.com/watch?v=ta_ZVS7HkwI`\n- Downloads the video as mp3, and adds metadata it detects.\n\n`vidl mp3 https://www.youtube.com/watch?v=ta_ZVS7HkwI --no-md`\n- Downloads the video as mp3, without adding metadata.\n\n`vidl`\n- Prints vidl\'s help menu, which looks like this:\n    ```\n    Download Usage:\n        vidl [format] [options] <URL>\n\n    Download Options:\n        format             mp3, mp4, wav or m4a. Default mp3\n        --no-md            Don\'t add metadata to downloaded files\n        --no-smart-md      Don\'t extract artist and song name from title\n        --no-embed         Don\'t embed thumbnail\n        --no-dl            Don\'t download anything. Usually used with -v\n        -v, --verbose      Display all logs\n\n    General Options:\n        --version          Show version. vidl -v and vidl version works too\n        -h, --help         Show this help message. vidl help works too\n        --config-path      Show the location of the configuration file\n\n    Update:\n        pip3 install vidl --upgrade --upgrade-strategy eager\n    ```\n\n# Configuration\nvidl has a configuration file, which you can find the location of by running `vidl --config-path`. In it, you can set the download folder and filename template.\n\nIf you screw something up, you can delete the file, and the default will be recreated the next time you run vidl.\n\n# Custom metadata parsing\n\nvidl has a `user_md_parser.py` file. By default, it does nothing, but you can configure it to manipulate metadata of songs you download however you like. In my case, I set the "Comment" metadata to "NCS" if the title ends with "[NCS Release]".\n\nDocumentation for this can be found in the file itself. The file is in the same folder as vidl\'s config file, which you can find by by running `vidl --config-path`. If you screw something up, you can delete the file, and the default will be recreated the next time you run vidl.\n\n# <a name="macos-shortcut-setup"></a>Set up shortcuts for vidl (macOS)\nYou\'ll be able to select any piece of text, press your chosen shortcut and the link(s) in your selected text will be downloaded! A little tedious to set up, but well worth it.\n\nFirst, we need to create a macOS Service:\n1. Open the Automator app.\n2. Choose File > New, and select Service.\n3. (TLDR; Add `Run Shell Script`) In the window that just popped up, there are two columns on the left (if not, click the `Library` button in the status bar). Select `Utilities` in the first column, and in the second column, drag `Run Shell Script` into the main part of the window.\n4. Make your settings match these:\n\n    ![Service receives selected [URLs] in [any application]. Input is [only URLs]. In your Run Shell Script box; Shell: [/bin/bash]. Pass input: [as arguments]](https://raw.githubusercontent.com/SpectralKH/vidl/master/macos-service-screenshot.png)\n\n    If you want the shortcut to only work in one app, select that app instead of `any application`.\n5. In the text box in the "Run Shell Script" box, paste in the following script:\n    ```bash\n    for f in "$@"\n    do\n      # behave like a normal terminal window:\n      export PATH=/usr/local/bin:$PATH\n      source ~/.bash_profile\n\n      vidl "$f"\n    done\n    ```\n6. Choose File > Save. Type in vidl.\n\nAlmost done, you just need to tie a shortcut to the macOS Service you just created:\n1. Open the System Preferences app.\n2. Go to Keyboard and select the Shortcuts tab.\n3. Select Services from the left column, and locate vidl (should be under Internet). Add your preferred shortcut.\n\n# API Usage\nYou can use Python to download a URL. Example:\n```python\nfrom vidl import config, dl\n\ndef main():\n    config.verify_config()\n    dl.download({\n        # all options are required\n        \'url\': \'https://www.youtube.com/watch?v=82IZ63TU8Fs\'\n        \'file_format\': \'mp3\',\n        \'audio_only\': True,\n        \'no_md\': False,\n        \'no_thumbnail_embed\': False,\n        \'no_smart_md\': False,\n        \'no_dl\': False,\n        \'verbose\': False,\n        \'download_folder\': config.get_config(\'download_folder\'),\n        \'output_template\': config.get_config(\'output_template\'),\n    })\nmain()\n```\n\n# Dev Instructions\n\n## Dev Installation\n1. Install Python. You may want to install it using [pyenv](https://github.com/pyenv/pyenv) in order to manage Python versions (If Poetry doesn\'t detect the right version, you can fix it with pyenv).\n2. Install [ffmpeg and ffprobe](https://www.ffmpeg.org/)\n3. Install [Poetry](https://poetry.eustace.io)\n4. Run `poetry install` to install Python package dependencies.\n5. If you\'re not on macOS or Windows, you need to specify where vidl will download files to in your vidl config file. Run `vidl --config-path` to see where the config file is.\n\nI recommend running `poetry config virtualenvs.in-project true`, which makes Poetry store your Python virtual environment inside the project folder. Additionally, it lets VSCode\'s Python extension detect the virtual environment if you set the `python.pythonPath` setting to `${workspaceFolder}/.venv/bin/python` in your settings.\n\n## Running\n```\npoetry run vidl\n```\n\nAn alternative to `poetry run <command>` is to run `poetry shell` to enter the virtual environment\'s Bash CLI, and then run the command on it\'s own (e.g `vidl`).\n\n## Releasing a new version\n\nFirst of all, consider updating the lockfile dependencies by running `poetry update`, then check if things still work.\n\n1. Bump the version number. `<version>` can be `patch`, `minor`, `major` or a version number:\n    ```\n    poetry version <version>\n    ```\n2. Update CHANGELOG.md\n3. Build:\n    ```\n    poetry build\n    ```\n4. Commit and create a git tag\n5. Create GitHub release with release notes and attach the build files\n6. Publish to PyPI:\n    ```\n    poetry publish\n    ```\n',
    'author': 'kasper.space',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/probablykasper/vidl',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
